///
///	@file mpv.cpp		@brief A mpv player plugin for VDR. forked from vdr-play
///
///	Copyright (c) 2012, 2013 by Johns.  All Rights Reserved.
///
///	Contributor(s): Dennis Bendlin
///
///	License: AGPLv3
///
///	This program is free software: you can redistribute it and/or modify
///	it under the terms of the GNU Affero General Public License as
///	published by the Free Software Foundation, either version 3 of the
///	License.
///
///	This program is distributed in the hope that it will be useful,
///	but WITHOUT ANY WARRANTY; without even the implied warranty of
///	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
///	GNU Affero General Public License for more details.
///
///	$Id: 2703e55d67db22d387428f995aeff440c18f86f7 $
//////////////////////////////////////////////////////////////////////////////

#include <vdr/interface.h>
#include <vdr/plugin.h>
#include <vdr/player.h>
#include <vdr/osd.h>
#include <vdr/shutdown.h>
#include <vdr/status.h>
#include <vdr/videodir.h>
#include <vdr/recording.h>

#include <mpv/client.h>

#ifdef HAVE_CONFIG
#include "config.h"
#endif

#include "play_service.h"
extern "C"
{
#include "readdir.h"
#include "player.h"
}

//////////////////////////////////////////////////////////////////////////////

    /// vdr-plugin version number.
    /// Makefile extracts the version number for generating the file name
    /// for the distribution archive.
static const char *const VERSION = "0.0.2"
#ifdef GIT_REV
    "-GIT" GIT_REV
#endif
    ;

    /// vdr-plugin description.
static const char *const DESCRIPTION = trNOOP("mpv player plugin");

    /// vdr-plugin text of main menu entry
static const char *MAINMENUENTRY = trNOOP("MPV");

//////////////////////////////////////////////////////////////////////////////

static char ConfigHideMainMenuEntry;	///< hide main menu entry

static volatile int DoMakePrimary;	///< switch primary device to this

//////////////////////////////////////////////////////////////////////////////
//	cPlayer
//////////////////////////////////////////////////////////////////////////////

/**
**	Player class.
*/
class cMyPlayer:public cPlayer
{
  private:
    char *FileName;			///< file to play
    

  public:
     cMyPlayer(const char *);		///< player constructor
     virtual ~ cMyPlayer();		///< player destructor
    void Activate(bool);		///< player attached/detached
    /// get current replay mode
    virtual bool GetReplayMode(bool &, bool &, int &);
    //void UpdateMarks();
    mpv_handle *ctx;
};

/**
**	Player constructor.
**
**	@param filename	path and name of file to play
*/
cMyPlayer::cMyPlayer(const char *filename)
:cPlayer(pmExtern_THIS_SHOULD_BE_AVOIDED)
{
    dsyslog("[mpv]%s: '%s'\n", __FUNCTION__, filename);

    //PlayerTitle = new char[256];
/*    PlayerSetVolume(ctx, cDevice::CurrentVolume());
    dsyslog("[mpv]: initial volume %d\n", cDevice::CurrentVolume());
*/
    FileName = strdup(filename);
}

/**
**	Player destructor.
*/
cMyPlayer::~cMyPlayer()
{
    dsyslog("[mpv]%s: end\n", __FUNCTION__);

    PlayerStop();
    free(FileName);
    // FIXME: wait until primary device is switched?
    dsyslog("[mpv]: device %d->%d\n",
	cDevice::PrimaryDevice()->DeviceNumber(), DoMakePrimary);
}



/**
**	Player attached or detached.
**
**	@param on	flag turn player on or off
*/
void cMyPlayer::Activate(bool on)
{
    dsyslog("[mpv]%s: '%s' %d\n", __FUNCTION__, FileName, on);

    if (on) {
	ctx = PlayerStart(FileName);
    } else {
	PlayerStop();
    }
}

/**
**	Get current replay mode.
*/
bool cMyPlayer::GetReplayMode(bool & play, bool & forward, int &speed)
{
    play = !PlayerPaused;
    forward = true;
    if (PlayerSpeed == 1) {
	speed = -1;
    } else {
	speed = PlayerSpeed;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////
//	cStatus
//////////////////////////////////////////////////////////////////////////////

/**
**	Status class.
**
**	To get volume changes.
*/
class cMyStatus:public cStatus
{
  private:
    int Volume;				///< current volume

  public:
    cMyStatus(void);			///< my status constructor

  protected:
    virtual void SetVolume(int, bool);	///< volume changed
};

cMyStatus *Status;			///< status monitor for volume

/**
**	Status constructor.
*/
cMyStatus::cMyStatus(void)
{
    Volume = cDevice::CurrentVolume();

    dsyslog("[mpv]: status volume %d\n", Volume);
}

/**
**	Called if volume is set.
*/
void cMyStatus::SetVolume(int volume, bool absolute)
{
    dsyslog("[mpv]: volume %d %s\n", volume, absolute ? "abs" : "rel");

    if (absolute) {
	Volume = volume;
    } else {
	Volume += volume;
    }

    //PlayerSetVolume(ctx, Volume);
}

//////////////////////////////////////////////////////////////////////////////
//	cControl
//////////////////////////////////////////////////////////////////////////////
cMyPlayer * staticPlayer;
/**
**	Our player control class.
*/
class cMyControl:public cControl
{
  private:
    cMyPlayer * Player;			///< our player
    cSkinDisplayReplay *Display;	///< our osd display
    void ShowReplayMode(void);		///< display replay mode
    void ShowProgress(void);		///< display progress bar
    virtual void Show(void);		///< show replay control
    virtual void Hide(void);		///< hide replay control
    bool infoVisible;			///< RecordingInfo visible
    time_t timeoutShow;			///< timeout shown control
    int LastPlayerCurrent;
    cMarks marks;

  public:
    cMyControl(const char *);		///< player control constructor
    virtual ~ cMyControl();		///< player control destructor

    virtual eOSState ProcessKey(eKeys);	///< handle keyboard input
    cMarks *Marks() { return &marks; }
    void UpdateMarks();

};

void cMyControl::UpdateMarks()
{
 ((cList<cMark> *)&marks)->Clear();

 if (PlayerNumChapters > 1) {
 marks.Add(0);
 for (int i=0; i < PlayerNumChapters; i++) {
 	dsyslog("Setting Mark %d %d", i, PlayerChapters[i]/1000);
 marks.Add(PlayerChapters[i]/1000 - 1);/// 90000 * 25;//DEFAULTFRAMESPERSECOND; // assume 25fps ...
 marks.Add(PlayerChapters[i]/1000);/// 90000 * 25;//DEFAULTFRAMESPERSECOND; // assume 25fps ...
 }
 }
}

/**
**	Show replay mode.
*/
void cMyControl::ShowReplayMode(void)
{
    dsyslog("[mpv]%s: %d - %d\n", __FUNCTION__, Setup.ShowReplayMode,
	cOsd::IsOpen());

    // use vdr setup
    if (Display || (Setup.ShowReplayMode && !cOsd::IsOpen())) {
	bool play;
	bool forward;
	int speed;

	if (GetReplayMode(play, forward, speed)) {
	    if (!Display) {
		// no need to show normal play
		if (play && forward && speed == 1) {
		    return;
		}
		Display = Skins.Current()->DisplayReplay(true);
	    }
	    Display->SetMode(play, forward, speed);
	}
    }
}

/**
**	Show progress.
*/
void cMyControl::ShowProgress(void)
{
    if (Display || (!cOsd::IsOpen())) {
	bool play;
	bool forward;
	int speed;

  if (LastPlayerCurrent == PlayerCurrent)
    return;
  LastPlayerCurrent = PlayerCurrent; 

	if (GetReplayMode(play, forward, speed)) {
	    if (!Display) {
		Display = Skins.Current()->DisplayReplay(false);
	    }

	    if (!infoVisible) {
	    	UpdateMarks();
		infoVisible = true;
		timeoutShow = time(0) + Setup.ChannelInfoTime;
	    }
		char TitleBuffer[256];
		char TitleComplete[256];
	    if (strcmp(TitleOverride, "") != 0) {
	    	snprintf (TitleBuffer, sizeof(TitleBuffer), "%s", TitleOverride);
	    } else if (strcmp(PlayerTitle, "") != 0) {
		snprintf (TitleBuffer, sizeof(TitleBuffer), "%s", PlayerTitle);
	    } else {
		snprintf (TitleBuffer, sizeof(TitleBuffer), "%s", PlayerFilename);
	    }
	    if (PlayerNumChapters > 0)
	    {
	    	snprintf (TitleComplete, sizeof(TitleComplete), "%s (%d/%d)", TitleBuffer, PlayerChapter+1, PlayerNumChapters);
	    	Display->SetMarks(Marks());
	    }
	    else
	    {
	    	snprintf (TitleComplete, sizeof(TitleComplete), "%s", TitleBuffer);
	    }
	    Display->SetTitle(TitleComplete);
	    Display->SetProgress(PlayerCurrent, PlayerTotal);
	    Display->SetMode(play, forward, speed);
	    Display->SetCurrent(IndexToHMSF(PlayerCurrent, false, 1));
	    Display->SetTotal(IndexToHMSF(PlayerTotal, false, 1));
	}
	SetNeedsFastResponse(true);
	Skins.Flush();
    }
}

/**
**	Show control.
*/
void cMyControl::Show(void)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);
    if (Setup.ShowReplayMode)
	ShowReplayMode();
    else
	ShowProgress();
}

/**
**	Control constructor.
**
**	@param filename	pathname of file to play.
*/
cMyControl::cMyControl(const char *filename)
:cControl(Player = new cMyPlayer(filename))
{
	staticPlayer = Player;
    Display = NULL;
    Status = new cMyStatus;		// start monitoring volume
    infoVisible = false;

    //LastSkipKey = kNone;
    //LastSkipSeconds = REPLAYCONTROLSKIPSECONDS;
    //LastSkipTimeout.Set(0);
    cStatus::MsgReplaying(this, filename, filename, true);

    cDevice::PrimaryDevice()->ClrAvailableTracks(true);
}

/**
**	Control destructor.
*/
cMyControl::~cMyControl()
{
    dsyslog("[mpv]%s\n", __FUNCTION__);

    delete Player;

    //delete Display;
    delete Status;

    Hide();
    cStatus::MsgReplaying(this, NULL, NULL, false);
    //Stop();
}

/**
**	Hide control.
*/
void cMyControl::Hide(void)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);

    if (Display) {
	delete Display;

	Display = NULL;
	SetNeedsFastResponse(false);
    }
}

/**
**	Process keyboard input.
**
**	@param key	pressed or releaded key
*/
eOSState cMyControl::ProcessKey(eKeys key)
{
  eOSState state;

  if (key != kNone)
  {
    dsyslog("[mpv]%s: key=%d\n", __FUNCTION__, key);
  }

  if (!PlayerIsRunning()) 		// check if player is still alive
  {
    dsyslog("[mpv]: player died\n");
    Hide();
    //FIXME: Stop();
    cControl::Shutdown();
    return osEnd;
  }

  if (infoVisible) 			// if RecordingInfo visible then update
  {
    if (timeoutShow && time(0) > timeoutShow)
    {
      Hide();
      timeoutShow = 0;
      infoVisible = false;
    }
    else
      ShowProgress();
  }
  //state=cOsdMenu::ProcessKey(key);
  state = osContinue;
  switch ((int)key)			// cast to shutup g++ warnings
  {
    case kUp:
      if (PlayerDvdNav)
      {
        PlayerSendDvdNavUp(Player->ctx);
        break;
      }
    case kPlay:
      Hide();
      if (PlayerSpeed != 1)
      {
        PlayerSendSetSpeed(Player->ctx, PlayerSpeed = 1);
      }
      if (PlayerPaused)
      {
        PlayerSendPause(Player->ctx);
        PlayerPaused ^= 1;
      }
      Show();
    break;

    case kDown:
      if (PlayerDvdNav)
      {
	PlayerSendDiscNavDown(Player->ctx);
	break;
      }
    case kPause:
      PlayerSendPause(Player->ctx);
      PlayerPaused ^= 1;
      Show();
    break;

    case kFastRew | k_Release:
    case kLeft | k_Release:
/*    if (Setup.MultiSpeedMode)
    {
      break;
    }
    // FIXME:
    break;*/
    case kLeft:
      if (PlayerDvdNav)
      {
        PlayerSendDvdNavLeft(Player->ctx);
        break;
      }
    case kFastRew:
      if (PlayerSpeed > 1)
      {
        PlayerSendSetSpeed(Player->ctx, PlayerSpeed /= 2);
      } 
      else
      {
        PlayerSendSeek(Player->ctx, -10);
      }
      Show();
    break;
    case kRight:
      if (PlayerDvdNav)
      {
        PlayerSendDvdNavRight(Player->ctx);
	break;
      }
    case kFastFwd:
      if (PlayerSpeed < 32)
      {
        PlayerSendSetSpeed(Player->ctx, PlayerSpeed *= 2);
      }
      Show();
    break;

    case kRed:
      // FIXME: TimeSearch();
    break;
    case kGreen | k_Repeat:
    case kGreen:
      PlayerSendSeek(Player->ctx, -60);
    break;
    case kYellow | k_Repeat:
    case kYellow:
      PlayerSendSeek(Player->ctx, +60);
    break;
    case kStop:
    case kBlue:
      dsyslog("[mpv]: player stopped\n");
      Hide();
      // FIXME: Stop();
      PlayerSendQuit(Player->ctx);
      //cControl::Shutdown();
    break;
    //return osEnd;

    case kOk:
      if (PlayerDvdNav)
      {
        PlayerSendDiscNavSelect(Player->ctx);
	break;
      }
      if (infoVisible)
      {
        Hide();
        infoVisible = false;
      }
      else
        Show();
        break;

    case kBack:
      if (PlayerDvdNav)
      {
        PlayerSendDvdNavPrev(Player->ctx);
        break;
      }
      PlayerSendQuit(Player->ctx);
      // FIXME: need to select old directory and index
      // todo let us override this with the service interface
      cRemote::CallPlugin("play");
    return osBack;

    case kMenu:			// VDR: eats the keys
    case k5:
      PlayerSendDvdNavMenu(Player->ctx);
    break;

    case kAudio:			// VDR: eats the keys
    case k7:
      // FIXME: audio menu
      PlayerSendSwitchAudio(Player->ctx);
    break;
    case kSubtitles:		// VDR: eats the keys
    case k9:
      // FIXME: subtitle menu
      PlayerSendSubSelect(Player->ctx);
    break;
    case kNext:
      PlayerNextChapter(Player->ctx);
    break;
    case kPrev:
      PlayerPreviousChapter(Player->ctx);
    break;

    default:
      break;
  }
  return state;
}

/**
**	Play a file.
**
**	@param filename	path and file name
*/
static void PlayFile(const char *filename)
{
    dsyslog("[mpv]: play file '%s'\n", filename);
    cControl::Launch(new cMyControl(filename));
}

/**
**	Check if filename is an iso dvd image.
**
**	@param filename	path and file name
**
**	@retval	true	archive
**	@retval false	no archive
*/
static int IsIsoImage(const char *filename)
{

    /**
    **	Table of supported iso image suffixes.
    */
    static const NameFilter IsoFilters[] = {
#define FILTER(x) { sizeof(x) - 1, x }
	FILTER(".bin"),
	FILTER(".dvd"),
	FILTER(".img"),
	FILTER(".iso"),
	FILTER(".mdf"),
	FILTER(".nrg"),
#undef FILTER
	{0, NULL}
    };
    int i;
    int len;

    len = strlen(filename);
    for (i = 0; IsoFilters[i].String; ++i) {
	if (len >= IsoFilters[i].Length
	    && !strcasecmp(filename + len - IsoFilters[i].Length,
		IsoFilters[i].String)) {
	    return 1;
	}
    }
    return 0;
}

/**
**	Play a file, with type detection.
**
**	@param filename	path and file name
*/
static void PlayFileHandleType(const char *filename)
{
    if (IsIsoImage(filename)) {		// handle dvd iso images
	char *tmp;

	tmp = (char *)malloc(sizeof("dvd://menu/") + strlen(filename));
	stpcpy(stpcpy(tmp, "dvd://menu/"), filename);
	PlayFile(tmp);
	free(tmp);
	return;
    }
    PlayFile(filename);
}

//////////////////////////////////////////////////////////////////////////////
//	cOsdMenu
//////////////////////////////////////////////////////////////////////////////

static char ShowBrowser;		///< flag show browser
static const char *BrowserStartDir;	///< browser start directory
static const NameFilter *BrowserFilters;	///< browser name filters
static int DirStackSize;		///< size of directory stack
static int DirStackUsed;		///< entries used of directory stack
static char **DirStack;			///< current path directory stack

/**
**	Menu class.
*/
class cBrowser:public cOsdMenu
{
  private:
    const NameFilter *Filter;		///< current filter

    /// Create a browser menu for current directory
    void CreateMenu(void);
    /// Create a browser menu for new directory
    void NewDir(const char *, const NameFilter *);
    /// Handle menu level up
    eOSState LevelUp(void);
    /// Handle menu item selection
    eOSState Selected(void);

  public:
    /// File browser constructor
    cBrowser(const char *, const char *, const NameFilter *);
    /// File browser destructor
    virtual ~ cBrowser();
    /// Process keyboard input
    virtual eOSState ProcessKey(eKeys);
};

/**
**	Add item to menu.  Called from C.
**
**	@param obj	cBrowser object
**	@param text	menu text
*/
extern "C" void cBrowser__Add(void *obj, const char *text)
{
    cBrowser *menu;

    // fucking stupid C++ can't assign void* without warning:
    menu = (typeof(menu)) obj;
    menu->Add(new cOsdItem(text));
}

/**
**	Create browser directory menu.
*/
void cBrowser::CreateMenu(void)
{
    Clear();				// start with empty directory
    // FIXME: should show only directory name in title
    //SetTitle(DirStack[0]);
    Skins.Message(mtStatus, tr("Scanning directory..."));

    if (DirStackUsed > 1) {
	// FIXME: should show only path
	Add(new cOsdItem(DirStack[0]));
    }
    ReadDirectory(DirStack[0], 1, NULL, cBrowser__Add, this);
    ReadDirectory(DirStack[0], 0, Filter, cBrowser__Add, this);
    // FIXME: handle errors!

    Display();				// display build menu
    Skins.Message(mtStatus, NULL);	// clear read message
}

/**
**	Create directory menu.
**
**	@param path	directory path file name
**	@param filter	name selection filter
*/
void cBrowser::NewDir(const char *path, const NameFilter * filter)
{
    int n;
    char *pathname;

    n = strlen(path);
#if 1
    // FIXME: force caller to do
    if (path[n - 1] == '/') {		// force '/' terminated
	pathname = strdup(path);
    } else {
	pathname = (char *)malloc(n + 2);
	stpcpy(stpcpy(pathname, path), "/");
    }
#endif

    // push on directory stack
    if (DirStackUsed >= DirStackSize) {	// increase stack size
	DirStackSize = DirStackUsed + 1;
	DirStack =
	    (typeof(DirStack)) realloc(DirStack,
	    DirStackSize * sizeof(*DirStack));
    }
    memmove(DirStack + 1, DirStack, DirStackUsed * sizeof(*DirStack));
    DirStackUsed++;
    DirStack[0] = pathname;

    Filter = filter;

    CreateMenu();
}

/**
**	Menu constructor.
**
**	@param title	menu title
**	@param path	directory path file name
**	@param filter	name selection filter
*/
cBrowser::cBrowser(const char *title, const char *path,
    const NameFilter * filter)
:cOsdMenu(title)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);

    if (path) {				// clear stack, start new
	int i;

	// free the stored directory stack
	for (i = 0; i < DirStackUsed; ++i) {
	    free(DirStack[i]);
	}
	//free(DirStack);
	//DirStack = NULL;		// reuse the old stack
	DirStackUsed = 0;

	NewDir(path, filter);
	return;
    }

    Filter = filter;

    CreateMenu();
}

/**
**	Menu destructor.
*/
cBrowser::~cBrowser()
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);
}

/**
**	Handle level up.
*/
eOSState cBrowser::LevelUp(void)
{
    char *down;
    char *name;

    if (DirStackUsed <= 1) {		// top level reached
	return osEnd;
    }
    // go level up
    --DirStackUsed;
    down = DirStack[0];
    memmove(DirStack, DirStack + 1, DirStackUsed * sizeof(*DirStack));

    CreateMenu();

    // select item, where we gone down
    down[strlen(down) - 1] = '\0';	// remove trailing '/'
    name = strrchr(down, '/');
    if (name) {
	cOsdItem *item;
	const char *text;
	int i;

	for (i = 0; (item = Get(i)); ++i) {
	    text = item->Text();
	    if (!strcmp(text, name + 1)) {
		SetCurrent(item);
		// FIXME: Display already called!
		Display();		// display build menu
		break;
	    }
	}
    }

    free(down);

    return osContinue;
}

/**
**	Handle selected item.
*/
eOSState cBrowser::Selected(void)
{
    int current;
    const cOsdItem *item;
    const char *text;
    char *filename;
    char *tmp;

    current = Current();		// get current menu item index
    item = Get(current);
    text = item->Text();

    if (current == 0 && DirStackUsed > 1) {
	return LevelUp();
    }
    // +2: \0 + #
    filename = (char *)malloc(strlen(DirStack[0]) + strlen(text) + 2);
    // path is '/' terminated
    tmp = stpcpy(stpcpy(filename, DirStack[0]), text);
    if (!IsDirectory(filename)) {
	if (IsArchive(filename)) {	// handle archives
	    stpcpy(tmp, "#");
	    NewDir(filename, Filter);
	    free(filename);
	    // FIXME: if dir fails use keep old!
	    return osContinue;
	}
	PlayFileHandleType(filename);
	free(filename);
	return osEnd;
    }
    // handle DVD image
    if (!strcmp(text, "AUDIO_TS") || !strcmp(text, "VIDEO_TS")) {
	free(filename);
	tmp = (char *)malloc(sizeof("dvdnav:///") + strlen(DirStack[0]));
	strcpy(stpcpy(tmp, "dvdnav:///"), DirStack[0]);
	PlayFile(tmp);
	free(tmp);
	return osEnd;
    }
    stpcpy(tmp, "/");			// append '/'
    NewDir(filename, Filter);
    free(filename);
    // FIXME: if dir fails use keep old!
    return osContinue;
}

/**
**	Handle Menu key event.
**
**	@param key	key event
*/
eOSState cBrowser::ProcessKey(eKeys key)
{
    eOSState state;

    // call standard function
    state = cOsdMenu::ProcessKey(key);
    if (state || key != kNone) {
	dsyslog("[mpv]%s: state=%d key=%d\n", __FUNCTION__, state, key);
    }

    switch (state) {
	case osUnknown:
	    switch (key) {
		case kOk:
		    return Selected();
		case kBack:
		    return LevelUp();
		default:
		    break;
	    }
	    break;
	case osBack:
	    state = LevelUp();
	    if (state == osEnd) {	// top level reached
		ShowBrowser = 0;
		return osPlugin;
	    }
	default:
	    break;
    }
    return state;
}

//////////////////////////////////////////////////////////////////////////////
//	cOsdMenu
//////////////////////////////////////////////////////////////////////////////

/**
**	Play plugin menu class.
*/
class cPlayMenu:public cOsdMenu
{
  private:
  public:
    cPlayMenu(const char *, int = 0, int = 0, int = 0, int = 0, int = 0);
    virtual ~ cPlayMenu();
    virtual eOSState ProcessKey(eKeys);
};

/**
**	Play menu constructor.
*/
cPlayMenu::cPlayMenu(const char *title, int c0, int c1, int c2, int c3, int c4)
:cOsdMenu(title, c0, c1, c2, c3, c4)
{
    SetHasHotkeys();

    Add(new cOsdItem(hk(tr("Browse")), osUser1));
    Add(new cOsdItem(hk(tr("Play optical disc")), osUser2));
    Add(new cOsdItem(""));
    Add(new cOsdItem(""));
    Add(new cOsdItem(hk(tr("Play audio CD")), osUser5));
    Add(new cOsdItem(hk(tr("Play video DVD")), osUser6));
/*    Add(new cOsdItem(hk(tr("Browse audio")), osUser7));
    Add(new cOsdItem(hk(tr("Browse image")), osUser8));
    Add(new cOsdItem(hk(tr("Browse video")), osUser9));*/
}

/**
**	Play menu destructor.
*/
cPlayMenu::~cPlayMenu()
{
}

/**
**	Handle play plugin menu key event.
**
**	@param key	key event
*/
eOSState cPlayMenu::ProcessKey(eKeys key)
{
    eOSState state;

    if (key != kNone) {
	dsyslog("[mpv]%s: key=%d\n", __FUNCTION__, key);
    }
    // call standard function
    state = cOsdMenu::ProcessKey(key);

    switch (state) {
	case osUser1:
	    ShowBrowser = 1;
	    BrowserStartDir = ConfigBrowserRoot;
	    BrowserFilters = NULL;
	    return osPlugin;		// restart with OSD browser

	case osUser2:
	    Skins.Message(mtStatus,
		tr("Function not working yet, use 3 or 4"));
	    return osContinue;

	case osUser5:			// play audio cdrom
	    PlayFile("cdda://");
	    return osEnd;
	case osUser6:			// play dvd
	    PlayFile("dvdnav://");
	    return osEnd;
	default:
	    break;
    }
    return state;
}

//////////////////////////////////////////////////////////////////////////////
//	cOsd
//////////////////////////////////////////////////////////////////////////////

/**
**	My device plugin OSD class.
*/
class cMyOsd:public cOsd
{
  public:
  //	cMyPlayer * Player;			///< our player
    static volatile char Dirty;		///< flag force redraw everything
    int OsdLevel;			///< current osd level

    cMyOsd(int, int, uint);		///< osd constructor
    virtual ~ cMyOsd(void);		///< osd destructor
    virtual void Flush(void);		///< commits all data to the hardware
    virtual void SetActive(bool);	///< sets OSD to be the active one
};

volatile char cMyOsd::Dirty;		///< flag force redraw everything

/**
**	Sets this OSD to be the active one.
**
**	@param on	true on, false off
**
**	@note only needed as workaround for text2skin plugin with
**	undrawn areas.
*/
void cMyOsd::SetActive(bool on)
{
    dsyslog("[mpv]%s: %d\n", __FUNCTION__, on);

    if (Active() == on) {
	return;				// already active, no action
    }
    cOsd::SetActive(on);

    // ignore sub-title, if menu is open
    if (OsdLevel >= OSD_LEVEL_SUBTITLES && IsOpen()) {
	return;
    }

    if (on) {
	Dirty = 1;
	// only flush here if there are already bitmaps
	//if (GetBitmap(0)) {
	//    Flush();
	//}
	//OsdOpen();
    } else {
	OsdClose(staticPlayer->ctx);
    }
}

/**
**	Constructor OSD.
**
**	Initializes the OSD with the given coordinates.
**
**	@param left	x-coordinate of osd on display
**	@param top	y-coordinate of osd on display
**	@param level	level of the osd (smallest is shown)
*/
cMyOsd::cMyOsd(int left, int top, uint level)
:cOsd(left, top, level)
{
    /* FIXME: OsdWidth/OsdHeight not correct!
       dsyslog("[mpv]%s: %dx%d+%d+%d, %d\n", __FUNCTION__, OsdWidth(),
       OsdHeight(), left, top, level);
     */

    OsdLevel = level;
    SetActive(true);
}

/**
**	OSD Destructor.
**
**	Shuts down the OSD.
*/
cMyOsd::~cMyOsd(void)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);
    SetActive(false);
    // done by SetActive: OsdClose();
}

/**
**	Actually commits all data to the OSD hardware.
*/
void cMyOsd::Flush(void)
{
  cPixmapMemory *pm;

  if (!IsTrueColor())
  {
    return;
  }

  //dsyslog("[mpv]%s: level %d active %d\n", __FUNCTION__, OsdLevel, Active());

  if (!Active())			// this osd is not active
  {
    return;
  }
  // don't draw sub-title if menu is active
  if (OsdLevel >= OSD_LEVEL_SUBTITLES && IsOpen())
  {
    return;
  }
  //
  //	VDR draws subtitle without clearing the old
  //
  if (OsdLevel >= OSD_LEVEL_SUBTITLES)
  {
    OsdClear(staticPlayer->ctx);
    cMyOsd::Dirty = 1;
    dsyslog("[mpv]%s: subtitle clear\n", __FUNCTION__);
  }

  LOCK_PIXMAPS;
  while ((pm = dynamic_cast<cPixmapMemory *>(RenderPixmaps())))
  {
    int sw;
    int sh;	
    int x;
    int y;
    int w;
    int h;

    sw = Width() + Left();
    sh = Height()+ Top();
    x = Left() + pm->ViewPort().X();
    y = Top() + pm->ViewPort().Y();
    w = pm->ViewPort().Width();
    h = pm->ViewPort().Height();

    // dsyslog ("[mpv] OSD size %d %d\n", sw, sh);
    // dsyslog("[mpv]%s: draw %dx%d%+d%+d %p\n", __FUNCTION__, w, h, x, y, pm->Data());
    OsdDrawARGB(staticPlayer->ctx, sw, sh, x, y, w, h, pm->Data());

    DestroyPixmap(pm);
  }
  cMyOsd::Dirty = 0;
}

//////////////////////////////////////////////////////////////////////////////
//	cOsdProvider
//////////////////////////////////////////////////////////////////////////////

/**
**	My device plugin OSD provider class.
*/
class cMyOsdProvider:public cOsdProvider
{
  private:
    static cOsd *Osd;

  public:
    virtual cOsd * CreateOsd(int, int, uint);
    virtual bool ProvidesTrueColor(void);
    cMyOsdProvider(void);
};

cOsd *cMyOsdProvider::Osd;		///< single osd

/**
**	Create a new OSD.
**
**	@param left	x-coordinate of OSD
**	@param top	y-coordinate of OSD
**	@param level	layer level of OSD
*/
cOsd *cMyOsdProvider::CreateOsd(int left, int top, uint level)
{
    //dsyslog("[mpv]%s: %d, %d, %d\n", __FUNCTION__, left, top, level);

    return Osd = new cMyOsd(left, top, level);
}

/**
**	Check if this OSD provider is able to handle a true color OSD.
**
**	@returns true we are able to handle a true color OSD.
*/
bool cMyOsdProvider::ProvidesTrueColor(void)
{
    return true;
}

/**
**	Create cOsdProvider class.
*/
cMyOsdProvider::cMyOsdProvider(void)
:  cOsdProvider()
{
    //dsyslog("[mpv]%s:\n", __FUNCTION__);
}

//////////////////////////////////////////////////////////////////////////////
//	cMenuSetupPage
//////////////////////////////////////////////////////////////////////////////

/**
**	Play plugin menu setup page class.
*/
class cMyMenuSetupPage:public cMenuSetupPage
{
  protected:
    ///
    /// local copies of global setup variables:
    int HideMainMenuEntry;

    virtual void Store(void);

  public:
    cMyMenuSetupPage(void);
    virtual eOSState ProcessKey(eKeys);	// handle input
};

/**
**	Process key for setup menu.
*/
eOSState cMyMenuSetupPage::ProcessKey(eKeys key)
{
    eOSState state;

    state = cMenuSetupPage::ProcessKey(key);

    return state;
}

/**
**	Constructor setup menu.
**
**	Import global config variables into setup.
*/
cMyMenuSetupPage::cMyMenuSetupPage(void)
{
    HideMainMenuEntry = ConfigHideMainMenuEntry;

    Add(new cMenuEditBoolItem(tr("Hide main menu entry"), &HideMainMenuEntry,
	    trVDR("no"), trVDR("yes")));
}

/**
**	Store setup.
*/
void cMyMenuSetupPage::Store(void)
{
    SetupStore("HideMainMenuEntry", ConfigHideMainMenuEntry =
	HideMainMenuEntry);
}

//////////////////////////////////////////////////////////////////////////////
//	cDevice
//////////////////////////////////////////////////////////////////////////////

/**
**	Dummy device class.
*/
class cMyDevice:public cDevice
{
  public:
    cMyDevice(void);
    virtual ~ cMyDevice(void);

    virtual void GetOsdSize(int &, int &, double &);

  protected:
    virtual void MakePrimaryDevice(bool);
};

/**
**	Device constructor.
*/
cMyDevice::cMyDevice(void)
{
    //dsyslog("[mpv]%s\n", __FUNCTION__);
}

/**
**	Device destructor. (never called!)
*/
cMyDevice::~cMyDevice(void)
{
    //dsyslog("[mpv]%s:\n", __FUNCTION__);
}

/**
**	Informs a device that it will be the primary device.
**
**	@param on	flag if becoming or loosing primary
*/
void cMyDevice::MakePrimaryDevice(bool on)
{
    //dsyslog("[mpv]%s: %d\n", __FUNCTION__, on);

    cDevice::MakePrimaryDevice(on);
    if (on) {
	new cMyOsdProvider();
    }
}

/**
**	Returns the width, height and pixel_aspect ratio the OSD.
**
**	FIXME: Called every second, for nothing (no OSD displayed)?
*/
void cMyDevice::GetOsdSize(int &width, int &height, double &pixel_aspect)
{
    if (!&width || !&height || !&pixel_aspect) {
	esyslog(tr("[mpv]: GetOsdSize invalid pointer(s)\n"));
	return;
    }
    // TODO generic
    width=1920;
    height=1080;
    pixel_aspect = 16.0 / 9.0 / (double)width * (double)height;
}

//////////////////////////////////////////////////////////////////////////////
//	cPlugin
//////////////////////////////////////////////////////////////////////////////

static cMyDevice *MyDevice;		///< dummy device needed for osd

class cMyPlugin:public cPlugin
{
  public:
    cMyPlugin(void);
    virtual ~ cMyPlugin(void);
    virtual const char *Version(void);
    virtual const char *Description(void);
    virtual const char *CommandLineHelp(void);
    virtual bool ProcessArgs(int, char *[]);
    virtual bool Initialize(void);
    virtual void MainThreadHook(void);
    virtual const char *MainMenuEntry(void);
    virtual cOsdObject *MainMenuAction(void);
    virtual cMenuSetupPage *SetupMenu(void);
    virtual bool SetupParse(const char *, const char *);
    virtual bool Service(const char *, void * = NULL);
    virtual const char **SVDRPHelpPages(void);
    virtual cString SVDRPCommand(const char *, const char *, int &);
};

/**
**	Initialize any member variables here.
**
**	@note DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
**	VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
*/
cMyPlugin::cMyPlugin(void)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);
}

/**
**	Clean up after yourself!
*/
cMyPlugin::~cMyPlugin(void)
{
    dsyslog("[mpv]%s:\n", __FUNCTION__);
}

/**
**	Return plugin version number.
**
**	@returns version number as constant string.
*/
const char *cMyPlugin::Version(void)
{
    return VERSION;
}

/**
**	Return plugin short description.
**
**	@returns short description as constant string.
*/
const char *cMyPlugin::Description(void)
{
    return tr(DESCRIPTION);
}

/**
**	Return a string that describes all known command line options.
**
**	@returns command line help as constant string.
*/
const char *cMyPlugin::CommandLineHelp(void)
{
    return::CommandLineHelp();
}

/**
**	Process the command line arguments.
*/
bool cMyPlugin::ProcessArgs(int argc, char *argv[])
{
    return::ProcessArgs(argc, argv);
}

/**
**	Start any background activities the plugin shall perform.
*/
bool cMyPlugin::Initialize(void)
{
    //dsyslog("[mpv]%s:\n", __FUNCTION__);

    // FIXME: can delay until needed?
    //Status = new cMyStatus;		// start monitoring
    // FIXME: destructs memory

    MyDevice = new cMyDevice;
    return true;
}

/**
**	Create main menu entry.
*/
const char *cMyPlugin::MainMenuEntry(void)
{
    return ConfigHideMainMenuEntry ? NULL : tr(MAINMENUENTRY);
}

/**
**	Perform the action when selected from the main VDR menu.
*/
cOsdObject *cMyPlugin::MainMenuAction(void)
{
    //dsyslog("[mpv]%s:\n", __FUNCTION__);

    if (ShowBrowser) {
	const char *start;

	start = BrowserStartDir;	// root only one time
	BrowserStartDir = NULL;
	return new cBrowser("Browse", start, BrowserFilters);
    }
    return new cPlayMenu("Play");
}

/**
**	Receive requests or messages.
**
**	@param id	unique identification string that identifies the
**			service protocol
**	@param data	custom data structure
*/
bool cMyPlugin::Service(const char *id, void *data)
{
/*    if (strcmp(id, PLAY_OSD_3DMODE_SERVICE) == 0) {
	VideoSetOsd3DMode(0);
	Play_Osd3DModeService_v1_0_t *r =
	    (Play_Osd3DModeService_v1_0_t *) data;
	VideoSetOsd3DMode(r->Mode);
	return true;
    }*/
    if (strcmp(id, PLAY_START_PLAY_SERVICE) == 0) {
		Play_StartPlayService_v1_0_t *r = (Play_StartPlayService_v1_0_t *) data;
	/*	if (r->Title)
			snprintf (TitleOverride, sizeof(TitleOverride), "%s", r->Title);*/
		PlayFileHandleType(r->Filename);
		return true;
	}
	if (strcmp(id, PLAY_SET_TITLE_SERVICE) == 0) {
		Play_SetTitleService_v1_0_t *r = (Play_SetTitleService_v1_0_t *) data;
		snprintf (TitleOverride, sizeof(TitleOverride), "%s", r->Title);
	}
    return false;
}

/**
**	Return SVDRP commands help pages.
**
**	return a pointer to a list of help strings for all of the plugin's
**	SVDRP commands.
*/
const char **cMyPlugin::SVDRPHelpPages(void)
{
    static const char *HelpPages[] = {
	"3DOF\n" "	  TURN OFF 3D", "3DTB\n" "    TURN ON 3D TB",
	"3DSB\n" "	  TURN ON 3D SBS", NULL
    };
    return HelpPages;
}

/**
**	Handle SVDRP commands.
**
**	@param command		SVDRP command
**	@param option		all command arguments
**	@param reply_code	reply code
*/
cString cMyPlugin::SVDRPCommand(const char *command,
    __attribute__ ((unused)) const char *option,
    __attribute__ ((unused)) int &reply_code)
{
    if (!strcasecmp(command, "3DOF")) {
	//VideoSetOsd3DMode(0);
	return "3d off";
    }
    if (!strcasecmp(command, "3DSB")) {
	//VideoSetOsd3DMode(1);
	return "3d sbs";
    }
    if (!strcasecmp(command, "3DTB")) {
	//VideoSetOsd3DMode(2);
	return "3d tb";
    }
    return NULL;
}

/**
**	Called for every plugin once during every cycle of VDR's main program
**	loop.
*/
void cMyPlugin::MainThreadHook(void)
{
    // dsyslog("[mpv]%s:\n", __FUNCTION__);

    if (DoMakePrimary) {
	dsyslog("[mpv]: switching primary device to %d\n", DoMakePrimary);
	cDevice::SetPrimaryDevice(DoMakePrimary);
	DoMakePrimary = 0;
    }
}

/**
**	Return our setup menu.
*/
cMenuSetupPage *cMyPlugin::SetupMenu(void)
{
    //dsyslog("[mpv]%s:\n", __FUNCTION__);

    return new cMyMenuSetupPage;
}

/**
**	Parse setup parameters
**
**	@param name	paramter name (case sensetive)
**	@param value	value as string
**
**	@returns true if the parameter is supported.
*/
bool cMyPlugin::SetupParse(const char *name, const char *value)
{
    dsyslog("[mpv]%s: '%s' = '%s'\n", __FUNCTION__, name, value);

    if (!strcasecmp(name, "HideMainMenuEntry")) {
	ConfigHideMainMenuEntry = atoi(value);
	return true;
    }

    return false;
}

//////////////////////////////////////////////////////////////////////////////

int OldPrimaryDevice;			///< old primary device

/**
**	Enable dummy device.
*/
extern "C" void EnableDummyDevice(void)
{
    OldPrimaryDevice = cDevice::PrimaryDevice()->DeviceNumber() + 1;

    dsyslog("[mpv]: primary device %d to new %d\n", OldPrimaryDevice,
	MyDevice->DeviceNumber() + 1);

    //if (!cDevice::SetPrimaryDevice(MyDevice->DeviceNumber() + 1)) {
    DoMakePrimary = MyDevice->DeviceNumber() + 1;
    cOsdProvider::Shutdown();
    //}
}

/**
**	Disable dummy device.
*/
extern "C" void DisableDummyDevice(void)
{
    dsyslog("[mpv]: primary device %d to old %d\n",
	cDevice::PrimaryDevice()->DeviceNumber() + 1, OldPrimaryDevice);

    //if (!cDevice::SetPrimaryDevice(OldPrimaryDevice)) {
    DoMakePrimary = OldPrimaryDevice;
    OldPrimaryDevice = 0;
    cOsdProvider::Shutdown();
    //}
}

VDRPLUGINCREATOR(cMyPlugin);		// Don't touch this!
