///
///	@file player.c	@brief A play plugin for VDR.
///
///	Copyright (c) 2012, 2013 by Johns.  All Rights Reserved.
///
///	Contributor(s): Dennis Bendlin
///
///	License: AGPLv3
///
///	This program is free software: you can redistribute it and/or modify
///	it under the terms of the GNU Affero General Public License as
///	published by the Free Software Foundation, either version 3 of the
///	License.
///
///	This program is distributed in the hope that it will be useful,
///	but WITHOUT ANY WARRANTY; without even the implied warranty of
///	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
///	GNU Affero General Public License for more details.
///
///	$Id: 5bccffda753d5154497f822af3e6f06f64b055d3 $
//////////////////////////////////////////////////////////////////////////////

#include <sys/types.h>
#include <sys/wait.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <poll.h>
#include <pthread.h>

#include <libintl.h>
#define _(str) gettext(str)		///< gettext shortcut
#define _N(str) str			///< gettext_noop shortcut


#include <fcntl.h>

#ifdef USE_XRANDR
#include <X11/extensions/Xrandr.h>
#endif

#include <mpv/client.h>
#include <inttypes.h>

#include "player.h"
#include "misc.h"

//////////////////////////////////////////////////////////////////////////////

const char *ConfigBrowserRoot = "/";	///< browser starting point
static char ConfigRefreshRate;		///< refresh rate switching
static char *ConfigVideoOut;		///< video out device
static char *ConfigAudioOut = "alsa:device=default";            ///< audio out device
//static char *ConfigAudioMixer;		///< audio mixer device
//static char *ConfigMixerChannel;	///< audio mixer channel
static const char *ConfigX11Display = ":0.0";	///< x11 display
static const char *ConfigMplayerDevice = "/dev/dvd"; /// DVD-Drive for mpv
int running = 0;

uint8_t  osd_buffer[1920*1080*4];  // TODO generic

#define MPV_OBSERVE_TIME_POS 1
#define MPV_OBSERVE_DISC_MENU 2
#define MPV_OBSERVE_FPS 3
#define MPV_OBSERVE_FILENAME 4
#define MPV_OBSERVE_LENGTH 5
#define MPV_OBSERVE_CHAPTERS 6
#define MPV_OBSERVE_CHAPTER 7

// check mpv errors and send them to log
static inline void check_error(int status)
{
    if (status < 0) {
        Error("mpv API error: %s\n", mpv_error_string(status));
        exit(1);
    }
}

//////////////////////////////////////////////////////////////////////////////
//	Osd
//////////////////////////////////////////////////////////////////////////////

/**
**	Close OSD.
*/
void OsdClose(mpv_handle *ctx)
{
  unsigned int i;
  Debug(3, "mpv: %s\n", __FUNCTION__);

  SendCommand (ctx, "overlay_remove 1");
  for (i=0;i<sizeof(osd_buffer);i++)
    osd_buffer[i] = 0;
}

/**
**	Clear OSD.
*/
void OsdClear(mpv_handle *ctx)
{
  unsigned int i;
  Debug(3, "mpv: %s\n", __FUNCTION__);

  SendCommand (ctx, "overlay_remove 1\n");
  for (i=0;i<sizeof(osd_buffer);i++)
    osd_buffer[i] = 0;
}

/**
**	Draw osd pixmap
*/
void OsdDrawARGB(mpv_handle *ctx, int sw, int sh, int x, int y, int w, int h, const uint8_t * argb)
{
  int sx;
  int sy;
  int pos;
  char cmd[512];
  int fd = open("/tmp/play_osd_1", O_CREAT | O_RDWR, S_IWUSR | S_IRUSR);
  
  Debug(3, "mpv: %s %d,%d %d,%d\n", __FUNCTION__, x, y, w, h);

  for (sy = 0; sy < h; ++sy) {
    for (sx = 0; sx < w; ++sx) {
      pos=0; 
      pos = pos + ((sy+y)*sw*4);
      pos = pos + ((sx+x)*4);
      osd_buffer[pos + 0] = argb[(w * sy + sx) * 4 + 0];
      osd_buffer[pos + 1] = argb[(w * sy + sx) * 4 + 1];
      osd_buffer[pos + 2] = argb[(w * sy + sx) * 4 + 2];
      osd_buffer[pos + 3] = argb[(w * sy + sx) * 4 + 3];
    }
  }
  write (fd, osd_buffer, sizeof(osd_buffer));
  close(fd);
  snprintf (cmd, sizeof(cmd), "overlay_add 1 0 0 \"/tmp/play_osd_1\" 0 \"bgra\" %d %d %d\n", sw, sh, sw*4);
  SendCommand (ctx, cmd);
}

//////////////////////////////////////////////////////////////////////////////
//	External player
//////////////////////////////////////////////////////////////////////////////

static pid_t PlayerPid;			///< player pid
static pthread_t PlayerThread;		///< player decode thread
static int PlayerVolume = -1;		///< volume 0 - 100
int64_t PlayerDvdNav;			///< dvdnav active
char PlayerPaused;			///< player paused
char PlayerSpeed;			///< player playback speed
int PlayerCurrent;			///< current postion in seconds
int PlayerTotal;			///< total length in seconds
char PlayerTitle[256];			///< title from meta data
char *PlayerFilename;			///< filename
int PlayerFps;				///< frames per second
int OriginalFps=-1;			///< fps before switch
int PlayerNumChapters;
int PlayerChapter;
unsigned int PlayerChapters[100];
char TitleOverride[256];

/**
**	Change frame-rate.
**
**	@param target_rate	requested frame rate (24, 25, 50, 60, ...)
*/
static void ChangeFrameRate(int TargetRate)
{
#ifdef USE_XRANDR
  Display *Dpy;
  int RefreshRate;
  XRRScreenConfiguration *CurrInfo;

  if (!ConfigRefreshRate)
    return;
	
  if (TargetRate == 25)
    TargetRate = 50; // fix DVD audio and since this is doubled it's ok
	
  if (TargetRate == 23)
    TargetRate = 24;

  Dpy = XOpenDisplay(ConfigX11Display);

  if (Dpy)
  {
    short *Rates;
    int NumberOfRates;
    SizeID CurrentSizeId;
    Rotation CurrentRotation;
    int RateFound = 0;

    CurrInfo = XRRGetScreenInfo(Dpy, DefaultRootWindow(Dpy));
    RefreshRate = XRRConfigCurrentRate(CurrInfo);
    CurrentSizeId =
    XRRConfigCurrentConfiguration(CurrInfo, &CurrentRotation);
    Rates = XRRConfigRates(CurrInfo, CurrentSizeId, &NumberOfRates);

    while (NumberOfRates-- > 0)
    {
      if (TargetRate == *Rates++)
        RateFound = 1;
    }

    if ((RefreshRate != TargetRate) && (RateFound == 1))
    {
      OriginalFps = RefreshRate;
      XRRSetScreenConfigAndRate(Dpy, CurrInfo, DefaultRootWindow(Dpy),
      CurrentSizeId, CurrentRotation, TargetRate, CurrentTime);
    }

    XRRFreeScreenConfigInfo(CurrInfo);
    XCloseDisplay(Dpy);
  }
#endif
}

/**
**	Execute external player.
**
**	@param filename	path and name of file to play
*/
static void PlayerExec(mpv_handle *ctx, const char *filename)
{
  int64_t osdlevel = 0;
  const char *cmd[] = {"loadfile", filename, NULL};

  running = 1;
  check_error(mpv_set_option_string(ctx, "vo", "vdpau"));
  check_error(mpv_set_option_string(ctx, "hwdec", "vdpau"));
  check_error(mpv_set_option_string(ctx, "ao", ConfigAudioOut));
  check_error(mpv_set_option_string(ctx, "ad", "-spdif:mp3,-spdif:aac,spdif:*"));
  check_error(mpv_set_option_string(ctx, "cache", "no")); // video stutters if enabled
  check_error(mpv_set_option_string(ctx, "hwdec-codecs", "h264,vc1,wmv3,mpeg2video"));
  check_error(mpv_set_option_string(ctx, "slang", "deu,de,ger,eng,en")); // this can break menu display
  check_error(mpv_set_option_string(ctx, "alang", "deu,de,ger,eng,en"));
  check_error(mpv_set_option_string(ctx, "fullscreen", "yes"));
  check_error(mpv_set_option_string(ctx, "sub-forced-only", "yes"));
  check_error(mpv_set_option_string(ctx, "ontop", "yes"));
  check_error(mpv_set_option_string(ctx, "ad-spdif-dtshd", "yes"));
  check_error(mpv_set_option_string(ctx, "cursor-autohide", "always"));
  check_error(mpv_set_option(ctx, "osd-level", MPV_FORMAT_INT64, &osdlevel));
  
#ifdef DEBUG
  mpv_request_log_messages(ctx, "info");
#endif
  
  if (mpv_initialize(ctx) < 0)
  {
    Error ("mpv failed to initialize\n");
    
  }

  Debug (3, "start %s\n", filename);
  mpv_command(ctx, cmd);
  return;
}

/**
**	Execute external player.
**
**	@param filename	path and name of file to play
*/
static mpv_handle *PlayerForkAndExec(const char *filename)
{
  setlocale(LC_NUMERIC, "C"); // needed for newer libmpv versions
  mpv_handle *ctx = mpv_create();
  if (!ctx)
  {
    Error ("mpv: failed to create context\n");
    exit(-1); // does this brings us back to vdr?
  }
  
  PlayerExec(ctx, filename);
  return ctx;
}

void HandlePropertyChange(mpv_event *event)
{
  int64_t *mpv_int;
  char *mpv_string;
  mpv_event_property *property = (mpv_event_property *) event->data;
  
  if (!property->data)
    return;
  
  // don't display on time-pos change since this floods the log
  if (event->reply_userdata != MPV_OBSERVE_TIME_POS)
  {
    Debug (3, "event %s \n", property->name);
  }
  
  switch (event->reply_userdata)
  {
    case MPV_OBSERVE_TIME_POS :
      mpv_string = *(char **)(property->data);
      PlayerCurrent = atoi(mpv_string);
    break;
  	
    case MPV_OBSERVE_DISC_MENU :
      mpv_int = *(int64_t **)(property->data);
      PlayerDvdNav = (int)mpv_int;
    break;
  	
    case MPV_OBSERVE_FPS :
      mpv_string = *(char **)(property->data);
      PlayerFps = atoi(mpv_string);
    break;
  	
    case MPV_OBSERVE_FILENAME :
      mpv_string = *(char **)(property->data);
      PlayerFilename = mpv_string;
    break;
  	
    case MPV_OBSERVE_LENGTH :
      mpv_string = *(char **)(property->data);
      PlayerTotal = atoi(mpv_string);
    break;
  	
    case MPV_OBSERVE_CHAPTERS :
      mpv_int = *(int64_t **)(property->data);
      PlayerNumChapters = (int)mpv_int;
    break;
  	
    case MPV_OBSERVE_CHAPTER :
      mpv_int = *(int64_t **)(property->data);
      PlayerChapter = (int)mpv_int;
    break;
  }
}

//////////////////////////////////////////////////////////////////////////////
//	Thread
//////////////////////////////////////////////////////////////////////////////

/**
**	External player handler thread.
**
**	@param dummy	dummy pointer
*/
static void *PlayerHandlerThread(void *handle)
{
  mpv_event_end_file *end_file;
  mpv_handle *ctx = (mpv_handle*) handle;
  struct mpv_event_log_message *msg;

  // set properties which should be observed
  mpv_observe_property(ctx, MPV_OBSERVE_TIME_POS, "time-pos", MPV_FORMAT_STRING);
  mpv_observe_property(ctx, MPV_OBSERVE_DISC_MENU, "disc-menu-active", MPV_FORMAT_FLAG);
  mpv_observe_property(ctx, MPV_OBSERVE_FPS, "fps", MPV_FORMAT_STRING);
  mpv_observe_property(ctx, MPV_OBSERVE_FILENAME, "filename", MPV_FORMAT_STRING);
  mpv_observe_property(ctx, MPV_OBSERVE_LENGTH, "length", MPV_FORMAT_STRING);
  mpv_observe_property(ctx, MPV_OBSERVE_CHAPTERS, "chapters", MPV_FORMAT_INT64);
  mpv_observe_property(ctx, MPV_OBSERVE_CHAPTER, "chapter", MPV_FORMAT_INT64);
    
  Debug(3, "mpv: player thread started\n");
  while (running)
  {
    mpv_event *event = mpv_wait_event(ctx, 10000);
    switch (event->event_id)
    {
      case MPV_EVENT_SHUTDOWN :
        running = 0;
        mpv_detach_destroy(ctx);
      break;
      
      case MPV_EVENT_END_FILE :
        end_file = (mpv_event_end_file *) event->data;
        if ( end_file->reason==0)
        	SendCommand(ctx, "quit");
      break;

     	case MPV_EVENT_PROPERTY_CHANGE :
        if (running)
          HandlePropertyChange(event);
     	break;

     	case MPV_EVENT_PLAYBACK_RESTART :
          ChangeFrameRate(PlayerFps); // switching directly after the fps event causes black screen
      break;
      
      case MPV_EVENT_LOG_MESSAGE :
      	msg = (struct mpv_event_log_message *)event->data;
      	Debug (3, msg->text);
      break;
        		
   		case MPV_EVENT_VIDEO_RECONFIG :
      case MPV_EVENT_NONE :
      case MPV_EVENT_GET_PROPERTY_REPLY :
      case MPV_EVENT_SET_PROPERTY_REPLY :
      case MPV_EVENT_COMMAND_REPLY :
      case MPV_EVENT_START_FILE :
      case MPV_EVENT_FILE_LOADED :
      case MPV_EVENT_TRACKS_CHANGED :
      case MPV_EVENT_TRACK_SWITCHED :
      case MPV_EVENT_IDLE :
      case MPV_EVENT_PAUSE :
      case MPV_EVENT_UNPAUSE :
      case MPV_EVENT_TICK :
      case MPV_EVENT_SCRIPT_INPUT_DISPATCH :
      case MPV_EVENT_CLIENT_MESSAGE :
      case MPV_EVENT_AUDIO_RECONFIG :
      case MPV_EVENT_METADATA_UPDATE :
      case MPV_EVENT_SEEK :
      case MPV_EVENT_CHAPTER_CHANGE :
      default :
      	Debug(3,"event: %d %s\n", event->event_id, mpv_event_name(event->event_id));
      break;
    }
  }
  
  return handle;
}

/**
**	Initialize player threads.
*/
static void PlayerThreadInit(mpv_handle *ctx)
{
    pthread_create(&PlayerThread, NULL, PlayerHandlerThread, ctx);
}

/**
**	Exit and cleanup player threads.
*/
static void PlayerThreadExit(void)
{
    if (PlayerThread) {
	void *retval;

	Debug(3, "mpv: player thread canceled\n");
	if (pthread_cancel(PlayerThread)) {
	    Error(_("mpv: can't queue cancel player thread\n"));
	}
	if (pthread_join(PlayerThread, &retval) || retval != PTHREAD_CANCELED) {
	    Error(_("mpv: can't cancel player thread\n"));
	}

	if (ConfigRefreshRate)
	{
	    ChangeFrameRate(50);  //TODO make generic
	    OriginalFps = -1;
	}

	PlayerThread = 0;
    }
}

/**
**	Send command to player.
**
**	@param formst	printf format string
*/
static void SendCommand(mpv_handle *ctx, char *cmd, ...)
{
	va_list va;
  char buf[256];
  va_start(va, cmd);
  vsnprintf(buf, sizeof(buf), cmd, va);
  va_end(va);
	Debug(3, "sendcommand: %s\n", buf);
	mpv_command_string(ctx, buf);
}

/**
**	Send player quit.
*/
void PlayerSendQuit(mpv_handle *ctx)
{
	SendCommand(ctx, "quit");
}

/**
**	Send player pause.
*/
void PlayerSendPause(mpv_handle *ctx)
{
	SendCommand(ctx, "pause");
}

/**
**	Send player speed_set.
**
**	@param speed	mplayer speed
*/
void PlayerSendSetSpeed(mpv_handle *ctx, int speed)
{
	SendCommand(ctx, "set speed %d\n", speed);
}

/**
**	Send player seek.
**
**	@param seconds	seek in seconds relative to current position
*/
void PlayerSendSeek(mpv_handle *ctx, int seconds)
{
	SendCommand(ctx, "seek %d\n", seconds);
}

/**
**	Send player volume.
*/
void PlayerSendVolume(mpv_handle *ctx)
{
	// FIXME: %.2f could have a problem with LANG
//	SendCommand("pausing_keep volume %.2f 1\n",
	//    (PlayerVolume * 100.0) / 255);
}

/**
**	Send switch audio track.
*/
void PlayerSendSwitchAudio(mpv_handle *ctx)
{
	SendCommand(ctx, "cycle audio\n");
}

/**
**	Send select subtitle.
*/
void PlayerSendSubSelect(mpv_handle *ctx)
{
	SendCommand(ctx, "cycle sub\n");
}

/**
**	Send up for dvdnav.
*/
void PlayerSendDvdNavUp(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav up\n");
}

/**
**	Send down for dvdnav.
*/
void PlayerSendDiscNavDown(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav down");
}

/**
**	Send left for dvdnav.
*/
void PlayerSendDvdNavLeft(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav left\n");
}

/**
**	Send right for dvdnav.
*/
void PlayerSendDvdNavRight(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav right\n");
}

/**
**	Send select for dvdnav.
*/
void PlayerSendDiscNavSelect(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav select\n");
}

/**
**	Send prev for dvdnav.
*/
void PlayerSendDvdNavPrev(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav prev\n");
}

/**
**	Send menu for dvdnav.
*/
void PlayerSendDvdNavMenu(mpv_handle *ctx)
{
	SendCommand(ctx, "discnav menu\n");
}

/**
** Previous Chapter
*/
void PlayerPreviousChapter(mpv_handle *ctx)
{
   SendCommand(ctx, "cycle chapter -1\n");
}

/**
** Next Chapter
*/
void PlayerNextChapter(mpv_handle *ctx)
{
  SendCommand(ctx, "cycle chapter 1\n");
}

/**
**	Start external player.
**
**	@param filename	path and name of file to play
*/
mpv_handle *PlayerStart(const char *filename)
{
	mpv_handle *ctx;
  PlayerPid = 0;

  PlayerPaused = 0;
  PlayerSpeed = 1;
  PlayerDvdNav = 0;

  EnableDummyDevice();
  ctx = PlayerForkAndExec(filename);

	PlayerThreadInit(ctx);
  return ctx;
}

/**
**	Stop external player.
*/
void PlayerStop()
{
	int waittime;
	int timeout;
  PlayerThreadExit();
    
	waittime = 0;
	timeout = 500;			// 0.5s

	while (PlayerIsRunning() && waittime++ < timeout)
	{
	  usleep(1 * 1000);
	}
  PlayerPid = 0;
  DisableDummyDevice();
}

/**
**	Is external player still running?
*/
int PlayerIsRunning(void)
{
	return running;
}

/**
**	Set player volume.
**
**	@param volume	new volume (0..255)
*/
void PlayerSetVolume(mpv_handle *ctx, int volume)
{
    Debug(3, "player: set volume=%d\n", volume);

    if (PlayerVolume != volume) {
	PlayerVolume = volume;
	if (PlayerPid) {
	    PlayerSendVolume(ctx);
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//	Device/Plugin C part
//////////////////////////////////////////////////////////////////////////////

/**
**	Return command line help string.
*/
const char *CommandLineHelp(void)
{
    return "  -% device\tmpv dvd device\n"
	"  -/\t/dir\tbrowser root directory\n"
	"  -a audio\tmpv --ao (-a alsa:device=[hdmi:CARD=NVidia,DEV=3])\n"
#ifdef USE_XRANDR
	"  -r\t\tswitch modeline to refresh rate of played file\n"
#endif
	"  -v video\tmpv --vo (vdpau:deint=2:hqscaling=1)\n";
}

/**
**	Process the command line arguments.
**
**	@param argc	number of arguments
**	@param argv	arguments vector
*/
int ProcessArgs(int argc, char *const argv[])
{
  const char *s;

    //
    //	Parse arguments.
    //
#ifdef __FreeBSD__
    if (!strcmp(*argv, "play")) {
	++argv;
	--argc;
    }
#endif
  if ((s = getenv("DISPLAY")))
  {
	  ConfigX11Display = s;
  }
  else
    setenv("DISPLAY", ConfigX11Display, 1);

  for (;;)
  {
	  switch (getopt(argc, argv, "-%:/:a:rv:"))
	  {
	    case '%':			// dvd-device
		ConfigMplayerDevice = optarg;
		continue;
	    case '/':			// browser root
		ConfigBrowserRoot = optarg;
		continue;
	    case 'a':			// audio out
		ConfigAudioOut = optarg;
		continue;
	    case 'r':			// refresh rate
		ConfigRefreshRate = 1;
		continue;
	    case 'v':			// video out
		ConfigVideoOut = optarg;
		continue;
	    case EOF:
		break;
	    case '-':
		fprintf(stderr, _("We need no long options\n"));
		return 0;
	    case ':':
		fprintf(stderr, _("Missing argument for option '%c'\n"),
		    optopt);
		return 0;
	    default:
		fprintf(stderr, _("Unkown option '%c'\n"), optopt);
		return 0;
	}
	break;
    }
    while (optind < argc) {
	fprintf(stderr, _("Unhandled argument '%s'\n"), argv[optind++]);
    }

    return 1;
}
